/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file contains the necessary initialization of the ILI9341 display
*   controller and the necessary function to update a rectangular area of the
*   screen from an Embedded Wizard UI application.
*   The data transfer is done by using the hardware SPI controller via DMA.
*   The color format of the display is RGB565.
*
*   This driver assumes the following GPIO layout of the interface:
*   Control lines (SPI):
*     MISO - 25
*     MOSI - 23
*     CLK  - 19
*     /CS  - 22               => chip select, low active
*   Additional control lines:
*     DC   - 21               => register select (data/command)
*     /RST - 18               => reset of LCD module, low active
*     /BCKL - 5               => backlight, low active
*
*   This module is just a template - adapt pin configuration and timing (!)
*   according your hardware layout and hardware constraints.
*   Adjust device parameters, initialization and gamma correction values
*   according your display and your needs.
*   This driver was tested on a ESP32-WOVER-KIT display module (3,2" TFT 320x240).
*
*******************************************************************************/

#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_system.h"
#include "driver/spi_master.h"
#include "soc/gpio_struct.h"
#include "driver/gpio.h"

#include "ili9341_spi.h"

/* SPI/LCD GPIO pin definitions */
#define PIN_NUM_MISO                    25
#define PIN_NUM_MOSI                    23
#define PIN_NUM_CLK                     19
#define PIN_NUM_CS                      22

#define PIN_NUM_DC                      21
#define PIN_NUM_RST                     18
#define PIN_NUM_BCKL                     5

/* display specific settings */
#define DISPLAY_PIXEL_WIDTH             ((uint16_t)320)
#define DISPLAY_PIXEL_HEIGHT            ((uint16_t)240)

/* ILI9341 commands */
#define ILI9341_SWRESET                 0x01   /* Software Reset */
#define ILI9341_READ_DISPLAY_ID         0x04   /* Read display identification information */
#define ILI9341_RDDST                   0x09   /* Read Display Status */
#define ILI9341_RDDPM                   0x0A   /* Read Display Power Mode */
#define ILI9341_RDDMADCTL               0x0B   /* Read Display MADCTL */
#define ILI9341_RDDCOLMOD               0x0C   /* Read Display Pixel Format */
#define ILI9341_RDDIM                   0x0D   /* Read Display Image Format */
#define ILI9341_RDDSM                   0x0E   /* Read Display Signal Mode */
#define ILI9341_RDDSDR                  0x0F   /* Read Display Self-Diagnostic Result */
#define ILI9341_SPLIN                   0x10   /* Enter Sleep Mode */
#define ILI9341_SLEEP_OUT               0x11   /* Sleep out register */
#define ILI9341_PTLON                   0x12   /* Partial Mode ON */
#define ILI9341_NORMAL_MODE_ON          0x13   /* Normal Display Mode ON */
#define ILI9341_DINVOFF                 0x20   /* Display Inversion OFF */
#define ILI9341_DINVON                  0x21   /* Display Inversion ON */
#define ILI9341_GAMMA                   0x26   /* Gamma register */
#define ILI9341_DISPLAY_OFF             0x28   /* Display off register */
#define ILI9341_DISPLAY_ON              0x29   /* Display on register */
#define ILI9341_COLUMN_ADDR             0x2A   /* Colomn address register */
#define ILI9341_PAGE_ADDR               0x2B   /* Page address register */
#define ILI9341_GRAM                    0x2C   /* GRAM register */
#define ILI9341_RGBSET                  0x2D   /* Color SET */
#define ILI9341_RAMRD                   0x2E   /* Memory Read */
#define ILI9341_PLTAR                   0x30   /* Partial Area */
#define ILI9341_VSCRDEF                 0x33   /* Vertical Scrolling Definition */
#define ILI9341_TEOFF                   0x34   /* Tearing Effect Line OFF */
#define ILI9341_TEON                    0x35   /* Tearing Effect Line ON */
#define ILI9341_MAC                     0x36   /* Memory Access Control register*/
#define ILI9341_VSCRSADD                0x37   /* Vertical Scrolling Start Address */
#define ILI9341_IDMOFF                  0x38   /* Idle Mode OFF */
#define ILI9341_IDMON                   0x39   /* Idle Mode ON */
#define ILI9341_PIXEL_FORMAT            0x3A   /* Pixel Format register */
#define ILI9341_WRITE_MEM_CONTINUE      0x3C   /* Write Memory Continue */
#define ILI9341_READ_MEM_CONTINUE       0x3E   /* Read Memory Continue */
#define ILI9341_SET_TEAR_SCANLINE       0x44   /* Set Tear Scanline */
#define ILI9341_GET_SCANLINE            0x45   /* Get Scanline */
#define ILI9341_WDB                     0x51   /* Write Brightness Display register */
#define ILI9341_RDDISBV                 0x52   /* Read Display Brightness */
#define ILI9341_WCD                     0x53   /* Write Control Display register*/
#define ILI9341_RDCTRLD                 0x54   /* Read CTRL Display */
#define ILI9341_WRCABC                  0x55   /* Write Content Adaptive Brightness Control */
#define ILI9341_RDCABC                  0x56   /* Read Content Adaptive Brightness Control */
#define ILI9341_WRITE_CABC              0x5E   /* Write CABC Minimum Brightness */
#define ILI9341_READ_CABC               0x5F   /* Read CABC Minimum Brightness */
#define ILI9341_READ_ID1                0xDA   /* Read ID1 */
#define ILI9341_READ_ID2                0xDB   /* Read ID2 */
#define ILI9341_READ_ID3                0xDC   /* Read ID3 */
#define ILI9341_RGB_INTERFACE           0xB0   /* RGB Interface Signal Control */
#define ILI9341_FRMCTR1                 0xB1   /* Frame Rate Control (In Normal Mode) */
#define ILI9341_FRMCTR2                 0xB2   /* Frame Rate Control (In Idle Mode) */
#define ILI9341_FRMCTR3                 0xB3   /* Frame Rate Control (In Partial Mode) */
#define ILI9341_INVTR                   0xB4   /* Display Inversion Control */
#define ILI9341_BPC                     0xB5   /* Blanking Porch Control register */
#define ILI9341_DFC                     0xB6   /* Display Function Control register */
#define ILI9341_ETMOD                   0xB7   /* Entry Mode Set */
#define ILI9341_BACKLIGHT1              0xB8   /* Backlight Control 1 */
#define ILI9341_BACKLIGHT2              0xB9   /* Backlight Control 2 */
#define ILI9341_BACKLIGHT3              0xBA   /* Backlight Control 3 */
#define ILI9341_BACKLIGHT4              0xBB   /* Backlight Control 4 */
#define ILI9341_BACKLIGHT5              0xBC   /* Backlight Control 5 */
#define ILI9341_BACKLIGHT7              0xBE   /* Backlight Control 7 */
#define ILI9341_BACKLIGHT8              0xBF   /* Backlight Control 8 */
#define ILI9341_POWER1                  0xC0   /* Power Control 1 register */
#define ILI9341_POWER2                  0xC1   /* Power Control 2 register */
#define ILI9341_VCOM1                   0xC5   /* VCOM Control 1 register */
#define ILI9341_VCOM2                   0xC7   /* VCOM Control 2 register */
#define ILI9341_NVMWR                   0xD0   /* NV Memory Write */
#define ILI9341_NVMPKEY                 0xD1   /* NV Memory Protection Key */
#define ILI9341_RDNVM                   0xD2   /* NV Memory Status Read */
#define ILI9341_READ_ID4                0xD3   /* Read ID4 */
#define ILI9341_PGAMMA                  0xE0   /* Positive Gamma Correction register */
#define ILI9341_NGAMMA                  0xE1   /* Negative Gamma Correction register */
#define ILI9341_DGAMCTRL1               0xE2   /* Digital Gamma Control 1 */
#define ILI9341_DGAMCTRL2               0xE3   /* Digital Gamma Control 2 */
#define ILI9341_INTERFACE               0xF6   /* Interface control register */
#define ILI9341_POWERA                  0xCB   /* Power control A register */
#define ILI9341_POWERB                  0xCF   /* Power control B register */
#define ILI9341_DTCA                    0xE8   /* Driver timing control A */
#define ILI9341_DTCB                    0xEA   /* Driver timing control B */
#define ILI9341_POWER_SEQ               0xED   /* Power on sequence register */
#define ILI9341_3GAMMA_EN               0xF2   /* 3 Gamma enable register */
#define ILI9341_PRC                     0xF7   /* Pump ratio control register */


static spi_device_handle_t SpiDeviceHandle;
static spi_transaction_t   SpiTransaction[ 6 ];
static uint8_t             SpiTransferActive = 0;


/* helper function to set the D/C line according the flag provided within user field */
/* Attention: function is called in context of SPI interrupt */
void lcd_spi_pre_transfer_callback(spi_transaction_t *t)
{
  int dc = (int)t->user;
  gpio_set_level( PIN_NUM_DC, dc );
}


/* helper function to write a command register value in polling mode */
static void ILI9341_WriteCmnd( uint8_t aCmnd )
{
  esp_err_t        ret;
  spi_transaction_t t;

  /* clear entire transaction buffer */
  memset( &t, 0, sizeof( t ));

  /* initialize transaction buffer */
  t.length       = 8;                     /* commands are always one byte */
  t.tx_data[ 0 ] = aCmnd;                 /* store the command */
  t.flags        = SPI_TRANS_USE_TXDATA;

  /* send the command in polling mode */
  ret = spi_device_transmit( SpiDeviceHandle, &t );
  assert( ret == ESP_OK );
}


/* helper function to read a data byte in polling mode */
static uint8_t ILI9341_ReadByte( void )
{
  esp_err_t         ret;
  spi_transaction_t t;

  /* clear entire transaction buffer */
  memset( &t, 0, sizeof( t ));

  /* initialize transaction buffer */
  t.length       = 8;                     /* read one byte */
  t.flags        = SPI_TRANS_USE_RXDATA;
  t.user         = (void*)1;              /* set user member to 1 to indicate data transfer */

  /* send the data byte in polling mode */
  ret = spi_device_transmit( SpiDeviceHandle, &t );
  assert( ret == ESP_OK );

  return t.rx_data[ 0 ];
}


/* helper function to write a data byte in polling mode */
static void ILI9341_WriteByte( uint8_t aData )
{
  esp_err_t         ret;
  spi_transaction_t t;

  /* clear entire transaction buffer */
  memset( &t, 0, sizeof( t ));

  /* initialize transaction buffer */
  t.length       = 8;                     /* write one byte */
  t.tx_data[ 0 ] = aData;                 /* store the data byte */
  t.flags        = SPI_TRANS_USE_TXDATA;
  t.user         = (void*)1;              /* set user member to 1 to indicate data transfer */

  /* send the data byte in polling mode */
  ret = spi_device_transmit( SpiDeviceHandle, &t );
  assert( ret == ESP_OK );
}


/* helper function to initialize reset and backlight pins of LCD and to perform
   proper reset of display */
static void ILI9341_ResetDisplay()
{
  /* initialize non-SPI GPIOs */
  gpio_set_direction( PIN_NUM_DC, GPIO_MODE_OUTPUT );
  gpio_set_direction( PIN_NUM_RST, GPIO_MODE_OUTPUT );
  gpio_set_direction( PIN_NUM_BCKL, GPIO_MODE_OUTPUT );

  /* disable backlight */
  gpio_set_level( PIN_NUM_BCKL, 1 );

  /* apply hardware reset */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 5 / portTICK_PERIOD_MS );   /* reset signal asserted during 5ms  */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 20 / portTICK_PERIOD_MS );  /* reset signal asserted during 20ms */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_Init
*
* DESCRIPTION:
*   The function initializes the SPI hardware and all necessary GPIO in order
*   to reset and intialize the connected display hardware.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ILI9341_Init( void )
{
	esp_err_t                     ret;
  spi_bus_config_t              buscfg;
  spi_device_interface_config_t devcfg;

  /* clear configuration buffer */
  memset( &buscfg, 0, sizeof( buscfg ));
  memset( &devcfg, 0, sizeof( devcfg ));

  /* initialize the SPI bus configuration */
  buscfg.miso_io_num     = PIN_NUM_MISO;
  buscfg.mosi_io_num     = PIN_NUM_MOSI;
  buscfg.sclk_io_num     = PIN_NUM_CLK;
  buscfg.quadwp_io_num   = -1;
  buscfg.quadhd_io_num   = -1;
  buscfg.max_transfer_sz = DISPLAY_PIXEL_WIDTH * DISPLAY_PIXEL_HEIGHT * 2 + 8;

  ret = spi_bus_initialize( HSPI_HOST, &buscfg, 1 );
  ESP_ERROR_CHECK( ret );

  /* intialize the device interface configuration to communicate with ILI9341 */
#ifdef CONFIG_LCD_OVERCLOCK
  devcfg.clock_speed_hz = 26 * 1000 * 1000; /* 26 MHz - overclocked */
#else
  devcfg.clock_speed_hz = 10 * 1000 * 1000; /* 10 MHz - according spec */
#endif

  devcfg.mode         = 0;                             /* SPI mode 0 */
  devcfg.spics_io_num = PIN_NUM_CS;
  devcfg.queue_size   = 7;                             /* max number of transactions */
  devcfg.pre_cb       = lcd_spi_pre_transfer_callback; /* callback to handle D/C line */

  ret = spi_bus_add_device( HSPI_HOST, &devcfg, &SpiDeviceHandle );
  ESP_ERROR_CHECK( ret );

  /* perform reset of LCD */
  ILI9341_ResetDisplay();

  /* switch display off */
  ILI9341_WriteCmnd( ILI9341_DISPLAY_OFF );

  /* power control B */
  ILI9341_WriteCmnd( ILI9341_POWERB );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x83 ); // or 0xC1 or 0x81 ???
  ILI9341_WriteByte( 0x30 );

  /* power on sequence control */
  ILI9341_WriteCmnd( ILI9341_POWER_SEQ );
  ILI9341_WriteByte( 0x64 );
  ILI9341_WriteByte( 0x03 );
  ILI9341_WriteByte( 0x12 );
  ILI9341_WriteByte( 0x81 );

  /* driver timing control A */
  ILI9341_WriteCmnd( ILI9341_DTCA );
  ILI9341_WriteByte( 0x85 );
  ILI9341_WriteByte( 0x01 );  // or 0x00
  ILI9341_WriteByte( 0x79 );  // or 0x78

  /* power control A */
  ILI9341_WriteCmnd( ILI9341_POWERA );
  ILI9341_WriteByte( 0x39 );
  ILI9341_WriteByte( 0x2C );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x34 );
  ILI9341_WriteByte( 0x02 );

  /* pump ratio control */
  ILI9341_WriteCmnd( ILI9341_PRC );
  ILI9341_WriteByte( 0x20 );

  /* driver timing control B */
  ILI9341_WriteCmnd( ILI9341_DTCB );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );

  /* power control 1 */
  ILI9341_WriteCmnd( ILI9341_POWER1 );
  ILI9341_WriteByte( 0x26 );    // = 4.75 V

  /* power control 2 */
  ILI9341_WriteCmnd( ILI9341_POWER2 );
  ILI9341_WriteByte( 0x11 );  // or 0x10

  /* VCOM 1 / 2 registers */
  ILI9341_WriteCmnd( ILI9341_VCOM1 );
  ILI9341_WriteByte( 0x35 );  // or 0x45
  ILI9341_WriteByte( 0x3E );  // or 0x15
  ILI9341_WriteCmnd( ILI9341_VCOM2 );
  ILI9341_WriteByte( 0xBE );  // or 0x90

  /* memory access control */
  ILI9341_WriteCmnd( ILI9341_MAC );
  ILI9341_WriteByte( 0x28 ); // or 0x48 or 0xC8 ???

  /* pixel format set */
  ILI9341_WriteCmnd( ILI9341_PIXEL_FORMAT );
  ILI9341_WriteByte( 0x55 );

  /* frame rate */
  ILI9341_WriteCmnd( ILI9341_FRMCTR1 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x1B );

  /* gamma configuration */
  ILI9341_WriteCmnd( ILI9341_3GAMMA_EN );
  ILI9341_WriteByte( 0x08 );

  ILI9341_WriteCmnd( ILI9341_GAMMA );
  ILI9341_WriteByte( 0x01 );

  ILI9341_WriteCmnd( ILI9341_PGAMMA );
  ILI9341_WriteByte( 0x1F );
  ILI9341_WriteByte( 0x1A );
  ILI9341_WriteByte( 0x18 );
  ILI9341_WriteByte( 0x0A );
  ILI9341_WriteByte( 0x0F );
  ILI9341_WriteByte( 0x06 );
  ILI9341_WriteByte( 0x45 );
  ILI9341_WriteByte( 0x87 );
  ILI9341_WriteByte( 0x32 );
  ILI9341_WriteByte( 0x0A );
  ILI9341_WriteByte( 0x07 );
  ILI9341_WriteByte( 0x02 );
  ILI9341_WriteByte( 0x07 );
  ILI9341_WriteByte( 0x05 );
  ILI9341_WriteByte( 0x00 );

  ILI9341_WriteCmnd( ILI9341_NGAMMA );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x25 );
  ILI9341_WriteByte( 0x27 );
  ILI9341_WriteByte( 0x05 );
  ILI9341_WriteByte( 0x10 );
  ILI9341_WriteByte( 0x09 );
  ILI9341_WriteByte( 0x3A );
  ILI9341_WriteByte( 0x78 );
  ILI9341_WriteByte( 0x4D );
  ILI9341_WriteByte( 0x05 );
  ILI9341_WriteByte( 0x18 );
  ILI9341_WriteByte( 0x0D );
  ILI9341_WriteByte( 0x38 );
  ILI9341_WriteByte( 0x3A );
  ILI9341_WriteByte( 0x1F );

  /* colomn address set */
  ILI9341_WriteCmnd( ILI9341_COLUMN_ADDR );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0xEF );

  /* page address set */
  ILI9341_WriteCmnd( ILI9341_PAGE_ADDR );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x00 );
  ILI9341_WriteByte( 0x01 );
  ILI9341_WriteByte( 0x3F );

  /* entry mode set */
  ILI9341_WriteCmnd( ILI9341_ETMOD );
  ILI9341_WriteByte( 0x07 );

  /* display function control set */
  ILI9341_WriteCmnd( ILI9341_DFC );
  ILI9341_WriteByte( 0x0A );
  ILI9341_WriteByte( 0x82 );
  ILI9341_WriteByte( 0x27 );
  ILI9341_WriteByte( 0x00 );

  /* wake up */
  ILI9341_WriteCmnd( ILI9341_SLEEP_OUT );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* switch display on */
  ILI9341_WriteCmnd( ILI9341_DISPLAY_ON );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* start writing to GRAM */
  ILI9341_WriteCmnd( ILI9341_GRAM );

  /* enable backlight */
  gpio_set_level( PIN_NUM_BCKL, 0 );
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_ReadID
*
* DESCRIPTION:
*   The function assumes a connected and intialized ILI9341 display and tries
*   to read its ID.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   If successful, the function returns the ID of the display - 0 otherwise.
*
*******************************************************************************/
uint16_t ILI9341_ReadID(void)
{
  uint16_t result;

  /* send command */
  ILI9341_WriteCmnd( ILI9341_READ_ID4 );

  /* read dummy data */
  ILI9341_ReadByte();

  /* read register values */
  ILI9341_ReadByte();
  result = ILI9341_ReadByte() << 8;
  result |= ILI9341_ReadByte();

  return result;
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_WaitForCompletion
*
* DESCRIPTION:
*   The function ensures that the previous SPI data transfer is completed..
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ILI9341_WaitForCompletion( void )
{
  if ( SpiTransferActive )
  {
    int                i;
  	spi_transaction_t* t;
    esp_err_t          ret;

    /* wait for all 6 transactions to be done */
    for ( i = 0; i < 6; i++ )
    {
      ret = spi_device_get_trans_result( SpiDeviceHandle, &t, portMAX_DELAY );
      assert( ret == ESP_OK );
    }

    /* clear the flag */
    SpiTransferActive = 0;
  }
}


/*******************************************************************************
* FUNCTION:
*   ILI9341_WriteDataWindow
*
* DESCRIPTION:
*   The function sets the destination position and size within the framebuffer
*   of the display according the given rectangle and starts the transfer of
*   the given data. The function does NOT wait for the completion of the transfer.
*
* ARGUMENTS:
*   aX       - Horizontal position for next write access.
*   aY       - Vertical position for next write access.
*   aW       - Width of the data window in pixel.
*   aH       - Height of the data window in pixel.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ILI9341_WriteDataWindow( uint16_t aX, uint16_t aY, uint16_t aW, uint16_t aH, uint8_t* aData )
{
  int       i;
  esp_err_t ret;
  int       xEnd = aX + aW - 1;
  int       yEnd = aY + aH - 1;

  if ( xEnd >= DISPLAY_PIXEL_WIDTH )
    xEnd = DISPLAY_PIXEL_WIDTH - 1;
  if ( yEnd >= DISPLAY_PIXEL_HEIGHT )
    yEnd = DISPLAY_PIXEL_HEIGHT - 1;

  /* ensure that previous update is completed */
  ILI9341_WaitForCompletion();

  /* clear entire transaction buffer */
  memset( &SpiTransaction[ 0 ], 0, sizeof( SpiTransaction ));

  /* colomn address set */
  SpiTransaction[ 0 ].tx_data[ 0 ] = ILI9341_COLUMN_ADDR;
  SpiTransaction[ 0 ].length       = 8;
  SpiTransaction[ 0 ].flags        = SPI_TRANS_USE_TXDATA;

  SpiTransaction[ 1 ].tx_data[ 0 ] = aX >> 8;     /* start column high byte */
  SpiTransaction[ 1 ].tx_data[ 1 ] = aX & 0xFF;   /* start column low byte */
  SpiTransaction[ 1 ].tx_data[ 2 ] = xEnd >> 8;   /* end column high byte */
  SpiTransaction[ 1 ].tx_data[ 3 ] = xEnd & 0xFF; /* end column low byte */
  SpiTransaction[ 1 ].length       = 4 * 8;
  SpiTransaction[ 1 ].user         = (void*)1;    /* indicate data transfer */
  SpiTransaction[ 1 ].flags        = SPI_TRANS_USE_TXDATA;

  /* page address set */
  SpiTransaction[ 2 ].tx_data[ 0 ] = ILI9341_PAGE_ADDR;
  SpiTransaction[ 2 ].length       = 8;
  SpiTransaction[ 2 ].flags        = SPI_TRANS_USE_TXDATA;

  SpiTransaction[ 3 ].tx_data[ 0 ] = aY >> 8;     /* start row high byte */
  SpiTransaction[ 3 ].tx_data[ 1 ] = aY & 0xFF;   /* start row low byte */
  SpiTransaction[ 3 ].tx_data[ 2 ] = yEnd >> 8;   /* end row high byte */
  SpiTransaction[ 3 ].tx_data[ 3 ] = yEnd & 0xFF; /* end row low byte */
  SpiTransaction[ 3 ].length       = 4 * 8;
  SpiTransaction[ 3 ].user         = (void*)1;    /* indicate data transfer */
  SpiTransaction[ 3 ].flags        = SPI_TRANS_USE_TXDATA;

  /* start writing to GRAM */
  SpiTransaction[ 4 ].tx_data[ 0 ] = ILI9341_GRAM;
  SpiTransaction[ 4 ].length       = 8;
  SpiTransaction[ 4 ].flags        = SPI_TRANS_USE_TXDATA;

  /* finally, send the data... */
  SpiTransaction[ 5 ].tx_buffer    = aData;
  SpiTransaction[ 5 ].length       = aW * aH * 2 * 8;
  SpiTransaction[ 5 ].user         = (void*)1;    /* indicate data transfer */

  /* queue all transactions */
  for ( i = 0; i < 6; i++ )
  {
    ret = spi_device_queue_trans( SpiDeviceHandle, &SpiTransaction[ i ], portMAX_DELAY );
    assert( ret == ESP_OK );
  }
  SpiTransferActive = 1;
}


/* msy */
