/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file contains the necessary initialization of the ST7789V display
*   controller and the necessary function to update a rectangular area of the
*   screen from an Embedded Wizard UI application.
*   The data transfer is done by using the hardware SPI controller via DMA.
*   The color format of the display is RGB565.
*
*   This driver assumes the following GPIO layout of the interface:
*   Control lines (SPI):
*     MISO - 25
*     MOSI - 23
*     CLK  - 19
*     /CS  - 22               => chip select, low active
*   Additional control lines:
*     DC   - 21               => register select (data/command)
*     /RST - 18               => reset of LCD module, low active
*     /BCKL - 5               => backlight, low active
*
*   This module is just a template - adapt pin configuration and timing (!)
*   according your hardware layout and hardware constraints.
*   Adjust device parameters, initialization and gamma correction values
*   according your display and your needs.
*   This driver was tested on a ESP32-WOVER-KIT display module (3,2" TFT 320x240).
*
*******************************************************************************/

#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_system.h"
#include "driver/spi_master.h"
#include "soc/gpio_struct.h"
#include "driver/gpio.h"

#include "st7789v_spi.h"

/* SPI/LCD GPIO pin definitions */
#define PIN_NUM_MISO                    25
#define PIN_NUM_MOSI                    23
#define PIN_NUM_CLK                     19
#define PIN_NUM_CS                      22

#define PIN_NUM_DC                      21
#define PIN_NUM_RST                     18
#define PIN_NUM_BCKL                     5

/* display specific settings */
#define DISPLAY_PIXEL_WIDTH             ((uint16_t)320)
#define DISPLAY_PIXEL_HEIGHT            ((uint16_t)240)

/* ST7789V commands */
#define ST7789V_SWRESET                 0x01   /* Software Reset */
#define ST7789V_RDDID                   0x04   /* Read Display ID Information */
#define ST7789V_RDDST                   0x09   /* Read Display Status */
#define ST7789V_RDDPM                   0x0A   /* Read Display Power Mode */
#define ST7789V_RDDMADCTL               0x0B   /* Read Display MADCTL */
#define ST7789V_RDDCOLMOD               0x0C   /* Read Display Pixel Format */
#define ST7789V_RDDIM                   0x0D   /* Read Display Image Format */
#define ST7789V_RDDSM                   0x0E   /* Read Display Signal Mode */
#define ST7789V_RDDSDR                  0x0F   /* Read Display Self-Diagnostic Result */
#define ST7789V_SPLIN                   0x10   /* Enter Sleep Mode */
#define ST7789V_SLPOUT                  0x11   /* Sleep out */
#define ST7789V_PTLON                   0x12   /* Partial Mode ON */
#define ST7789V_NORON                   0x13   /* Normal Display Mode ON */
#define ST7789V_INVOFF                  0x20   /* Display Inversion OFF */
#define ST7789V_INVON                   0x21   /* Display Inversion ON */
#define ST7789V_GAMSET                  0x26   /* Gamma register */
#define ST7789V_DISPOFF                 0x28   /* Display off */
#define ST7789V_DISPON                  0x29   /* Display on */
#define ST7789V_CASET                   0x2A   /* Colomn address */
#define ST7789V_RASET                   0x2B   /* Row address */
#define ST7789V_RAMWR                   0x2C   /* Memory Write */
#define ST7789V_RAMRD                   0x2E   /* Memory Read */
#define ST7789V_PLTAR                   0x30   /* Partial Area */
#define ST7789V_VSCRDEF                 0x33   /* Vertical Scrolling Definition */
#define ST7789V_TEOFF                   0x34   /* Tearing Effect Line OFF */
#define ST7789V_TEON                    0x35   /* Tearing Effect Line ON */
#define ST7789V_MADCTL                  0x36   /* Memory Access Control */
#define ST7789V_VSCSAD                  0x37   /* Vertical Scrolling Start Address */
#define ST7789V_IDMOFF                  0x38   /* Idle Mode OFF */
#define ST7789V_IDMON                   0x39   /* Idle Mode ON */
#define ST7789V_COLMOD                  0x3A   /* Pixel Format register */
#define ST7789V_WRMEMC                  0x3C   /* Write Memory Continue */
#define ST7789V_RDMEMC                  0x3E   /* Read Memory Continue */
#define ST7789V_STE                     0x44   /* Set Tear Scanline */
#define ST7789V_GSCAN                   0x45   /* Get Scanline */
#define ST7789V_WRDISBV                 0x51   /* Write Brightness Display */
#define ST7789V_RDDISBV                 0x52   /* Read Display Brightness */
#define ST7789V_WRCTRLD                 0x53   /* Write Control Display */
#define ST7789V_RDCTRLD                 0x54   /* Read CTRL Display */
#define ST7789V_WRCACE                  0x55   /* Write Content Adaptive Brightness Control */
#define ST7789V_RDCABC                  0x56   /* Read Content Adaptive Brightness Control */
#define ST7789V_WRCABCMB                0x5E   /* Write CABC Minimum Brightness */
#define ST7789V_RDCABCMB                0x5F   /* Read CABC Minimum Brightness */
#define ST7789V_RDABCSDR                0x68   /* Read Automatic Brightness Control Self Diagnostic Result */
#define ST7789V_READ_ID1                0xDA   /* Read ID1 */
#define ST7789V_READ_ID2                0xDB   /* Read ID2 */
#define ST7789V_READ_ID3                0xDC   /* Read ID3 */
#define ST7789V_RAMCTRL                 0xB0   /* RAM Control */
#define ST7789V_RGBCTRL                 0xB1   /* RGB Interface Control */
#define ST7789V_PORCTRL                 0xB2   /* Porch Setting */
#define ST7789V_FRCTRL1                 0xB3   /* Frame Rate Control (In Partial Mode) */
#define ST7789V_PARCTRL                 0xB5   /* Partial Mode Control */
#define ST7789V_GCTRL                   0xB7   /* Gate Control */
#define ST7789V_GTADJ                   0xB8   /* Gate On Timing Adjustment */
#define ST7789V_DGMEN                   0xBA   /* Digital Gamma Enable */
#define ST7789V_VCOMS                   0xBB   /* VCOMS Settings */
#define ST7789V_LCMCTRL                 0xC0   /* LCM Control */
#define ST7789V_IDSET                   0xC1   /* ID Code Setting */
#define ST7789V_VDVVRHEN                0xC2   /* VDV and VRH Command Enable */
#define ST7789V_VRHS                    0xC3   /* VRH Set */
#define ST7789V_VDVS                    0xC4   /* VDV Set */
#define ST7789V_VCMOFSET                0xC5   /* VCOMS Offset Set */
#define ST7789V_FRCTRL2                 0xC6   /* Frame Rate Control in Normal Mode */
#define ST7789V_CABCCTRL                0xC7   /* CABC Control */
#define ST7789V_REGSEL1                 0xC8   /* Register Value Selection 1 */
#define ST7789V_REGSEL2                 0xCA   /* Register Value Selection 2 */
#define ST7789V_PWMFRSEL                0xCC   /* PWM Frequency Selection */
#define ST7789V_PWCTRL1                 0xD0   /* Power Control 1 */
#define ST7789V_VAPVANEN                0xD2   /* Enable VAP/VAN singal output */
#define ST7789V_CMD2EN                  0xDF   /* Command 2 Enable */
#define ST7789V_PVGAMCTRL               0xE0   /* Positive Gamma Correction */
#define ST7789V_NVGAMCTRL               0xE1   /* Negative Gamma Correction */
#define ST7789V_DGMLUTR                 0xE2   /* Digital Gamma LUT Red */
#define ST7789V_DGMLUTB                 0xE3   /* Digital Gamma LUT Blue */
#define ST7789V_GATECTRL                0xE4   /* Gate Control */
#define ST7789V_SPI2EN                  0xE7   /* SPI2 Enable */
#define ST7789V_PWCTRL2                 0xE8   /* Power Control 2 */
#define ST7789V_EQCTRL                  0xE9   /* Equalize time Control */
#define ST7789V_PROMCTRL                0xEC   /* Program Mode Control */
#define ST7789V_PROMEN                  0xFA   /* Program Mode Enable */
#define ST7789V_NVMSET                  0xFC   /* NVM Setting */
#define ST7789V_PROMACT                 0xFE   /* Program Action */


static spi_device_handle_t SpiDeviceHandle;
static spi_transaction_t   SpiTransaction[ 6 ];
static uint8_t             SpiTransferActive = 0;


/* helper function to set the D/C line according the flag provided within user field */
/* Attention: function is called in context of SPI interrupt */
void lcd_spi_pre_transfer_callback(spi_transaction_t *t)
{
  int dc = (int)t->user;
  gpio_set_level( PIN_NUM_DC, dc );
}


/* helper function to write a command register value in polling mode */
static void ST7789V_WriteCmnd( uint8_t aCmnd )
{
  esp_err_t        ret;
  spi_transaction_t t;

  /* clear entire transaction buffer */
  memset( &t, 0, sizeof( t ));

  /* initialize transaction buffer */
  t.length       = 8;                     /* commands are always one byte */
  t.tx_data[ 0 ] = aCmnd;                 /* store the command */
  t.flags        = SPI_TRANS_USE_TXDATA;

  /* send the command in polling mode */
  ret = spi_device_transmit( SpiDeviceHandle, &t );
  assert( ret == ESP_OK );
}


/* helper function to read a data byte in polling mode */
static uint8_t ST7789V_ReadByte( void )
{
  esp_err_t         ret;
  spi_transaction_t t;

  /* clear entire transaction buffer */
  memset( &t, 0, sizeof( t ));

  /* initialize transaction buffer */
  t.length       = 8;                     /* read one byte */
  t.flags        = SPI_TRANS_USE_RXDATA;
  t.user         = (void*)1;              /* set user member to 1 to indicate data transfer */

  /* send the data byte in polling mode */
  ret = spi_device_transmit( SpiDeviceHandle, &t );
  assert( ret == ESP_OK );

  return t.rx_data[ 0 ];
}


/* helper function to write a data byte in polling mode */
static void ST7789V_WriteByte( uint8_t aData )
{
  esp_err_t         ret;
  spi_transaction_t t;

  /* clear entire transaction buffer */
  memset( &t, 0, sizeof( t ));

  /* initialize transaction buffer */
  t.length       = 8;                     /* write one byte */
  t.tx_data[ 0 ] = aData;                 /* store the data byte */
  t.flags        = SPI_TRANS_USE_TXDATA;
  t.user         = (void*)1;              /* set user member to 1 to indicate data transfer */

  /* send the data byte in polling mode */
  ret = spi_device_transmit( SpiDeviceHandle, &t );
  assert( ret == ESP_OK );
}


/* helper function to initialize reset and backlight pins of LCD and to perform
   proper reset of display */
static void ST7789V_ResetDisplay()
{
  /* initialize non-SPI GPIOs */
  gpio_set_direction( PIN_NUM_DC, GPIO_MODE_OUTPUT );
  gpio_set_direction( PIN_NUM_RST, GPIO_MODE_OUTPUT );
  gpio_set_direction( PIN_NUM_BCKL, GPIO_MODE_OUTPUT );

  /* disable backlight */
  gpio_set_level( PIN_NUM_BCKL, 1 );

  /* apply hardware reset */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 5 / portTICK_PERIOD_MS );   /* reset signal asserted during 5ms  */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
  gpio_set_level( PIN_NUM_RST, 0 );
  vTaskDelay( 20 / portTICK_PERIOD_MS );  /* reset signal asserted during 20ms */
  gpio_set_level( PIN_NUM_RST, 1 );
  vTaskDelay( 10 / portTICK_PERIOD_MS );  /* reset signal released during 10ms */
}


/*******************************************************************************
* FUNCTION:
*   ST7789V_Init
*
* DESCRIPTION:
*   The function initializes the SPI hardware and all necessary GPIO in order
*   to reset and intialize the connected display hardware.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ST7789V_Init( void )
{
	esp_err_t                     ret;
  spi_bus_config_t              buscfg;
  spi_device_interface_config_t devcfg;

  /* clear configuration buffer */
  memset( &buscfg, 0, sizeof( buscfg ));
  memset( &devcfg, 0, sizeof( devcfg ));

  /* initialize the SPI bus configuration */
  buscfg.miso_io_num     = PIN_NUM_MISO;
  buscfg.mosi_io_num     = PIN_NUM_MOSI;
  buscfg.sclk_io_num     = PIN_NUM_CLK;
  buscfg.quadwp_io_num   = -1;
  buscfg.quadhd_io_num   = -1;
  buscfg.max_transfer_sz = DISPLAY_PIXEL_WIDTH * DISPLAY_PIXEL_HEIGHT * 2 + 8;

  ret = spi_bus_initialize( HSPI_HOST, &buscfg, 1 );
  ESP_ERROR_CHECK( ret );

  /* intialize the device interface configuration to communicate with ST7789V */
#ifdef CONFIG_LCD_OVERCLOCK
  devcfg.clock_speed_hz = 26 * 1000 * 1000; /* 26 MHz - overclocked */
#else
  devcfg.clock_speed_hz = 10 * 1000 * 1000; /* 10 MHz - according spec */
#endif

  devcfg.mode         = 0;                             /* SPI mode 0 */
  devcfg.spics_io_num = PIN_NUM_CS;
  devcfg.queue_size   = 7;                             /* max number of transactions */
  devcfg.pre_cb       = lcd_spi_pre_transfer_callback; /* callback to handle D/C line */

  ret = spi_bus_add_device( HSPI_HOST, &devcfg, &SpiDeviceHandle );
  ESP_ERROR_CHECK( ret );

  /* perform reset of LCD */
  ST7789V_ResetDisplay();

  /* memory access control */
  ST7789V_WriteCmnd( ST7789V_MADCTL );
  ST7789V_WriteByte( 0x60 );

  /* interface pixel format, 16bits/pixel for RGB/MCU interface */
  ST7789V_WriteCmnd( ST7789V_COLMOD );
  ST7789V_WriteByte( 0x55 );

  /* porch setting */
  ST7789V_WriteCmnd( ST7789V_PORCTRL );
  ST7789V_WriteByte( 0x0C );
  ST7789V_WriteByte( 0x0C );
  ST7789V_WriteByte( 0x00 );
  ST7789V_WriteByte( 0x33 );
  ST7789V_WriteByte( 0x33 );

  /* gate control, Vgh=13.65V, Vgl=-10.43V */
  ST7789V_WriteCmnd( ST7789V_GCTRL );
  ST7789V_WriteByte( 0x45 );

  /* VCOM Setting, VCOM=1.175V */
  ST7789V_WriteCmnd( ST7789V_VCOMS );
  ST7789V_WriteByte( 0x2B );

  /* LCM Control, XOR: BGR, MX, MH */
  ST7789V_WriteCmnd( ST7789V_LCMCTRL );
  ST7789V_WriteByte( 0x2C );

  /* VDV and VRH Command Enable, enable=1 */
  ST7789V_WriteCmnd( ST7789V_VDVVRHEN );
  ST7789V_WriteByte( 0x01 );
  ST7789V_WriteByte( 0xFF );

  /* VRH Set, Vap=4.4+... */
  ST7789V_WriteCmnd( ST7789V_VRHS );
  ST7789V_WriteByte( 0x11 );

  /* VDV Set, VDV=0 */
  ST7789V_WriteCmnd( ST7789V_VDVS );
  ST7789V_WriteByte( 0x20 );

  /* frame rate control, 60Hz, inversion=0 */
  ST7789V_WriteCmnd( ST7789V_FRCTRL2 );
  ST7789V_WriteByte( 0x0F );

  /* power control 1, AVDD=6.8V, AVCL=-4.8V, VDDS=2.3V */
  ST7789V_WriteCmnd( ST7789V_PWCTRL1 );
  ST7789V_WriteByte( 0xA4 );
  ST7789V_WriteByte( 0xA1 );

  /* positive voltage gamma control */
  ST7789V_WriteCmnd( ST7789V_PVGAMCTRL );
  ST7789V_WriteByte( 0xD0 );
  ST7789V_WriteByte( 0x00 );
  ST7789V_WriteByte( 0x05 );
  ST7789V_WriteByte( 0x0E );
  ST7789V_WriteByte( 0x15 );
  ST7789V_WriteByte( 0x0D );
  ST7789V_WriteByte( 0x37 );
  ST7789V_WriteByte( 0x43 );
  ST7789V_WriteByte( 0x47 );
  ST7789V_WriteByte( 0x09 );
  ST7789V_WriteByte( 0x15 );
  ST7789V_WriteByte( 0x12 );
  ST7789V_WriteByte( 0x16 );
  ST7789V_WriteByte( 0x19 );

  /* negative voltage gamma control */
  ST7789V_WriteCmnd( ST7789V_NVGAMCTRL );
  ST7789V_WriteByte( 0xD0 );
  ST7789V_WriteByte( 0x00 );
  ST7789V_WriteByte( 0x05 );
  ST7789V_WriteByte( 0x0D );
  ST7789V_WriteByte( 0x0C );
  ST7789V_WriteByte( 0x06 );
  ST7789V_WriteByte( 0x2D );
  ST7789V_WriteByte( 0x44 );
  ST7789V_WriteByte( 0x40 );
  ST7789V_WriteByte( 0x0E );
  ST7789V_WriteByte( 0x1C );
  ST7789V_WriteByte( 0x18 );
  ST7789V_WriteByte( 0x16 );
  ST7789V_WriteByte( 0x19 );

  /* sleep out */
  ST7789V_WriteCmnd( ST7789V_SLPOUT );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* display on */
  ST7789V_WriteCmnd( ST7789V_DISPON );
  vTaskDelay( 200 / portTICK_PERIOD_MS );

  /* enable backlight */
  gpio_set_level( PIN_NUM_BCKL, 0 );
}


/*******************************************************************************
* FUNCTION:
*   ST7789V_ReadID
*
* DESCRIPTION:
*   The function assumes a connected and intialized ST7789V display and tries
*   to read its ID.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   If successful, the function returns the ID of the display - 0 otherwise.
*
*******************************************************************************/
uint16_t ST7789V_ReadID(void)
{
  uint16_t result;

  /* send command */
  ST7789V_WriteCmnd( ST7789V_RDDID );

  /* read dummy data */
  ST7789V_ReadByte();

  /* read register values */
  ST7789V_ReadByte();
  result = ST7789V_ReadByte() << 8;
  result |= ST7789V_ReadByte();

  return result;
}


/*******************************************************************************
* FUNCTION:
*   ST7789V_WaitForCompletion
*
* DESCRIPTION:
*   The function ensures that the previous SPI data transfer is completed..
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ST7789V_WaitForCompletion( void )
{
  if ( SpiTransferActive )
  {
    int                i;
  	spi_transaction_t* t;
    esp_err_t          ret;

    /* wait for all 6 transactions to be done */
    for ( i = 0; i < 6; i++ )
    {
      ret = spi_device_get_trans_result( SpiDeviceHandle, &t, portMAX_DELAY );
      assert( ret == ESP_OK );
    }

    /* clear the flag */
    SpiTransferActive = 0;
  }
}


/*******************************************************************************
* FUNCTION:
*   ST7789V_WriteDataWindow
*
* DESCRIPTION:
*   The function sets the destination position and size within the framebuffer
*   of the display according the given rectangle and starts the transfer of
*   the given data. The function does NOT wait for the completion of the transfer.
*
* ARGUMENTS:
*   aX       - Horizontal position for next write access.
*   aY       - Vertical position for next write access.
*   aW       - Width of the data window in pixel.
*   aH       - Height of the data window in pixel.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void ST7789V_WriteDataWindow( uint16_t aX, uint16_t aY, uint16_t aW, uint16_t aH, uint8_t* aData )
{
  int       i;
  esp_err_t ret;
  int       xEnd = aX + aW - 1;
  int       yEnd = aY + aH - 1;

  if ( xEnd >= DISPLAY_PIXEL_WIDTH )
    xEnd = DISPLAY_PIXEL_WIDTH - 1;
  if ( yEnd >= DISPLAY_PIXEL_HEIGHT )
    yEnd = DISPLAY_PIXEL_HEIGHT - 1;

  /* ensure that previous update is completed */
  ST7789V_WaitForCompletion();

  /* clear entire transaction buffer */
  memset( &SpiTransaction[ 0 ], 0, sizeof( SpiTransaction ));

  /* colomn address set */
  SpiTransaction[ 0 ].tx_data[ 0 ] = ST7789V_CASET;
  SpiTransaction[ 0 ].length       = 8;
  SpiTransaction[ 0 ].flags        = SPI_TRANS_USE_TXDATA;

  SpiTransaction[ 1 ].tx_data[ 0 ] = aX >> 8;     /* start column high byte */
  SpiTransaction[ 1 ].tx_data[ 1 ] = aX & 0xFF;   /* start column low byte */
  SpiTransaction[ 1 ].tx_data[ 2 ] = xEnd >> 8;   /* end column high byte */
  SpiTransaction[ 1 ].tx_data[ 3 ] = xEnd & 0xFF; /* end column low byte */
  SpiTransaction[ 1 ].length       = 4 * 8;
  SpiTransaction[ 1 ].user         = (void*)1;    /* indicate data transfer */
  SpiTransaction[ 1 ].flags        = SPI_TRANS_USE_TXDATA;

  /* row address set */
  SpiTransaction[ 2 ].tx_data[ 0 ] = ST7789V_RASET;
  SpiTransaction[ 2 ].length       = 8;
  SpiTransaction[ 2 ].flags        = SPI_TRANS_USE_TXDATA;

  SpiTransaction[ 3 ].tx_data[ 0 ] = aY >> 8;     /* start row high byte */
  SpiTransaction[ 3 ].tx_data[ 1 ] = aY & 0xFF;   /* start row low byte */
  SpiTransaction[ 3 ].tx_data[ 2 ] = yEnd >> 8;   /* end row high byte */
  SpiTransaction[ 3 ].tx_data[ 3 ] = yEnd & 0xFF; /* end row low byte */
  SpiTransaction[ 3 ].length       = 4 * 8;
  SpiTransaction[ 3 ].user         = (void*)1;    /* indicate data transfer */
  SpiTransaction[ 3 ].flags        = SPI_TRANS_USE_TXDATA;

  /* start writing to GRAM */
  SpiTransaction[ 4 ].tx_data[ 0 ] = ST7789V_RAMWR;
  SpiTransaction[ 4 ].length       = 8;
  SpiTransaction[ 4 ].flags        = SPI_TRANS_USE_TXDATA;

  /* finally, send the data... */
  SpiTransaction[ 5 ].tx_buffer    = aData;
  SpiTransaction[ 5 ].length       = aW * aH * 2 * 8;
  SpiTransaction[ 5 ].user         = (void*)1;    /* indicate data transfer */

  /* queue all transactions */
  for ( i = 0; i < 6; i++ )
  {
    ret = spi_device_queue_trans( SpiDeviceHandle, &SpiTransaction[ i ], portMAX_DELAY );
    assert( ret == ESP_OK );
  }
  SpiTransferActive = 1;
}


/* msy */
